
# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san-web-client SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from flask_socketio import SocketIO

from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor


class AgentLogProcessor(MessageProcessor):
    """
    Tells the UI there's an agent message to process.
    """

    def __init__(self, socketio: SocketIO, sid: str):
        """
        Constructor
        """
        self.socketio: SocketIO = socketio
        self.sid: str = sid

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message:
          - Log the message
          - Highlight the agent in the network diagram
          - Display the message in the Agents Communication panel
        :param chat_message_dict: The chat message
        :param message_type: The type of message
        """
        if message_type != ChatMessageType.AGENT:
            # This is a framework message that contains chat context, system prompts or consolidated messages.
            # Don't log them. And there's no agent to highlight in the network diagram.
            return

        # Discard empty messages
        log_text = chat_message_dict.get("text", "").strip()
        if log_text == "":
            return

        # We have a valid message. Process it.
        # Get the list of agents that participated in the message and the name of the last agent,
        # which is the one that sent the message.
        origin_chain, last_agent_name = self.get_origin(chat_message_dict)

        header = "---------- ChatMessage ----------"
        log_message = f"{header}\n{origin_chain}: {log_text}"
        # Log the message
        print(log_message)
        # Highlight the last agent name in the network diagram
        # and display the log message in the Agents Communication panel
        self.display_message_in_ui(last_agent_name, log_message)

    @staticmethod
    def get_origin(chat_message_dict) -> Tuple[str, str]:
        """
        Get the list of agents that lead to this message and the agent that produced it.
        :param chat_message_dict: The chat message
        :return: A tuple containing:
        - the list of agents involved, as a string,
        - the name of the agent that produced the message
        """
        origin: List = chat_message_dict.get("origin")
        # Convert the origin list to a string
        origin_str = " -> ".join(tool['tool'] for tool in origin)
        # the agent name is the last tool in the origin list
        last_agent_name = origin[-1].get("tool")

        return origin_str, last_agent_name

    def display_message_in_ui(self, agent_name: str, message_log: str) -> None:
        """
        Displays the agent message in the UI
        Sends the agent name and its log to the UI to:
         - Highlight the agent node in the network diagram
         - Display the agent log in the Agents Communication panel

        """
        log_dict = {
            "agent_name": agent_name,
            "log": message_log
        }
        self.socketio.emit('agent_log', log_dict, room=self.sid)
        # Allow the event loop to process and send WebSocket messages before continuing execution.
        self.socketio.sleep(0)



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san-web-client SDK Software in commercial settings.
#
import argparse
import os
import threading
from typing import Any
from typing import Dict

from flask import Flask
from flask import redirect
from flask import render_template
from flask import request
from flask import session
from flask import url_for
from flask_socketio import SocketIO

from neuro_san.client.streaming_input_processor import StreamingInputProcessor
from neuro_san.session.service_agent_session import GrpcServiceAgentSession

from neuro_san_web_client.agent_log_processor import AgentLogProcessor

# Initialize a lock
user_sessions_lock = threading.Lock()
user_sessions = {}

app = Flask(__name__)
app.secret_key = 'your_secret_key'  # Replace with a secure key
socketio = SocketIO(app, async_mode='eventlet')

# Default configuration
DEFAULT_CONFIG = {
    'server_host': 'localhost',
    'server_port': 30011,
    'web_client_port': 5001,
    'agent_name': 'telco_network_support',
    'thinking_file': '/tmp/agent_thinking.txt',
    'thinking_dir': '/tmp'
}


@app.route('/', methods=['GET', 'POST'])
def index():
    if request.method == 'POST':
        # Update configuration based on user input
        session['server_host'] = request.form.get('host', app.config.get('server_host'))
        session['server_port'] = int(request.form.get('port', app.config.get('server_port')))
        session['agent_name'] = request.form.get('agent_name', app.config.get('agent_name'))
        # Initialize agent session with new config
        session['agent_session'] = None
        # Redirect to the index page to avoid form resubmission messages on refresh
        return redirect(url_for('index'))

    return render_template('index.html',
                           agent_name=session.get('agent_name', app.config['agent_name']),
                           host=session.get('server_host', app.config['server_host']),
                           port=session.get('server_port', app.config['server_port']))


# noinspection PyUnresolvedReferences
@socketio.on('user_input')
def handle_user_input(data):
    user_input = data.get('message')

    # Get the Socket.IO session ID
    sid = request.sid

    # Retrieve or initialize user-specific data
    with user_sessions_lock:
        user_session = user_sessions.get(sid)
        if not user_session:
            # No session found: create a new one
            user_session = create_user_session(sid)
            user_sessions[sid] = user_session

        input_processor = user_session["input_processor"]
        state = user_session["state"]
        # Keep track of the sly_data. input_processor will wipe it out.
        sly_data = state["sly_data"]
        # Update user input in state
        state["user_input"] = user_input

        print("========== Processing user message ==========")
        # Calling the processor updates the state
        state = input_processor.process_once(state)

        # If any additional sly_data is returned, update it in the state
        returned_sly_data = state.get("returned_sly_data", None)
        if returned_sly_data:
            sly_data.update(returned_sly_data)
        state["sly_data"] = sly_data

        # This is now the users' new state
        user_session['state'] = state

        # Start a background task to display the agent's response
        last_chat_response = state.get("last_chat_response")
        socketio.start_background_task(target=background_response_handler, chat_response=last_chat_response, sid=sid)


def create_user_session(sid):
    host = session.get('server_host', app.config['server_host'])
    port = session.get('server_port', app.config['server_port'])
    agent_name = session.get('agent_name', app.config['agent_name'])
    agent_session = GrpcServiceAgentSession(
        host=host,
        port=port,
        agent_name=agent_name
    )
    input_processor = StreamingInputProcessor(default_input="",
                                              thinking_file=DEFAULT_CONFIG['thinking_file'],
                                              session=agent_session,
                                              thinking_dir=DEFAULT_CONFIG['thinking_dir'])
    # Add a processor to handle agent logs
    # and to highlight the agents that respond in the agent network diagram
    agent_log_processor = AgentLogProcessor(socketio, sid)
    input_processor.processor.add_processor(agent_log_processor)

    # Note: If nothing is specified the server assumes the chat_filter_type
    #       should be "MINIMAL", however for this client which is aimed at
    #       developers, we specifically want a default MAXIMAL client to
    #       show all the bells and whistles of the output that a typical
    #       end user will not care about and not appreciate the extra
    #       data charges on their cell phone.
    chat_filter: Dict[str, Any] = {
        "chat_filter_type": "MAXIMAL"
    }

    # Initialize the state for the user session
    state: Dict[str, Any] = {
        "last_chat_response": None,
        "num_input": 0,
        "chat_filter": chat_filter,
        "sly_data": {},
    }

    # Create the user session
    user_session = {
        'input_processor': input_processor,
        'state': state
    }
    return user_session


# noinspection PyUnresolvedReferences
@socketio.on('disconnect')
def handle_disconnect():
    sid = request.sid
    with user_sessions_lock:
        if sid in user_sessions:
            del user_sessions[sid]
    print(f"Client disconnected: {sid}")


def background_response_handler(chat_response: str, sid):
        socketio.emit('agent_response', {'message': chat_response}, room=sid)


def clear_thinking_file():
    # Clear out the previous thinking file
    #
    # Incorrectly flagged as destination of Path Traversal 5
    # Reason: thinking_file was previously checked with FileOfClass.check_file()
    #         which actually does the path traversal check. CheckMarx does not
    #         recognize pathlib as a valid library with which to resolve these kinds
    #         of issues.
    with open(DEFAULT_CONFIG['thinking_file'], "w", encoding="utf-8") as thinking:
        thinking.write("\n")


def parse_args():
    """
    Parses command-line arguments for server and agent configuration.
    Priority order:
    1. Command-line arguments (highest priority)
    2. Environment or local variables (medium priority)
    3. Default values from `DEFAULT_CONFIG` (fallback)
    """
    parser = argparse.ArgumentParser(description="Configure the Neuro SAN web client and server.")

    parser.add_argument('--server-host', type=str,
                        default=os.getenv("NEURO_SAN_SERVER_HOST", DEFAULT_CONFIG['server_host']),
                        help="Host address for the Neuro SAN server")
    parser.add_argument('--server-port', type=int,
                        default=int(os.getenv("NEURO_SAN_SERVER_PORT", DEFAULT_CONFIG['server_port'])),
                        help="Port number for the Neuro SAN server")
    parser.add_argument('--web-client-port', type=int,
                        default=int(os.getenv("NEURO_SAN_WEB_CLIENT_PORT", DEFAULT_CONFIG['web_client_port'])),
                        help="Port number for the web client")
    parser.add_argument('--agent-name', type=str,
                        default=os.getenv("NEURO_SAN_AGENT_NAME", DEFAULT_CONFIG['agent_name']),
                        help="Agent name for the session")

    args, _ = parser.parse_known_args()

    config = vars(args)

    print(f"Starting app with Configuration: {config}")
    return config

if __name__ == '__main__':
    a_config = parse_args()
    # Store config in Flask app for later use
    # Items can be accessed anywhere in Flask routes e.g. using app.config['agent_name']
    app.config.update(a_config)
    clear_thinking_file()
    # Start the app with the parsed configuration
    socketio.run(app, debug=True, allow_unsafe_werkzeug=True, port=a_config['web_client_port'])



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san-web-client SDK Software in commercial settings.
#
from pathlib import Path

import argparse
import os

from pyhocon import ConfigFactory
from pyvis.network import Network

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
PATH_TO_STATIC = os.path.join(ROOT_DIR, 'static')


class DiagramBuilder:
    """
    Builds a .html file containing an interactive agent network diagram
    from a .hocon file containing agent definitions and connections.
    """
    def __init__(self):
        self.args = None

    @staticmethod
    def parse_agent_definitions(agent_data):
        agent_graph = {}

        # Step 1: First pass to gather all agents and their details
        tools = agent_data.get("tools", [])
        for tool in tools:
            agent_name = tool.get("name", "unknown_agent")
            agent_info = {
                "instructions": tool.get("instructions", "No instructions"),
                "command": tool.get("command", "No command"),
                "class": tool.get("class", "No class")
            }
            agent_graph[agent_name] = {
                "info": agent_info,
                "connections": tool.get("tools", [])
            }

        return agent_graph

    @staticmethod
    def create_interactive_agent_graph(agent_graph, output_html):
        # Initialize a pyvis network graph
        net = Network(height="750px", width="100%", bgcolor="#222222",
                      font_color="white", directed=True)

        # Step 2: Keep track of which nodes have already been added
        existing_nodes = set()

        # Add all known agent nodes to the pyvis network
        for agent_name, agent_data in agent_graph.items():
            # Add node with hover information
            hover_text = (
                f"Instructions: {agent_data['info']['instructions']}<br>"
                f"Class: {agent_data['info']['class']}<br>"
                f"Command: {agent_data['info']['command']}<br>"
            )
            net.add_node(agent_name, title=hover_text, label=agent_name)
            existing_nodes.add(agent_name)

        # Step 3: Add all edges (and create missing nodes in a different color)
        for agent_name, agent_data in agent_graph.items():
            connections = agent_data["connections"]
            for connection in connections:
                if connection in agent_graph:
                    # Normal edge, since this connection/agent exists
                    net.add_edge(agent_name, connection)
                else:
                    # This tool/agent does not exist in agent_graph
                    print(f"Warning: {connection} referenced by {agent_name} does not exist.")

                    # Create a missing node if it hasn't been added yet
                    if connection not in existing_nodes:
                        net.add_node(connection, label=connection, color="green")
                        existing_nodes.add(connection)

                    # Add edge from agent_name to the newly created/missing node
                    net.add_edge(agent_name, connection)

        # Disable hierarchical layout and physics for free node movement
        net.set_options("""
         {
           "physics": {
             "enabled": false
           },
           "interaction": {
             "zoomView": false,
             "dragView": true
           },
           "manipulation": {
             "enabled": false
           },
           "autoResize": true
         }
         """)

        # Save the interactive network to an HTML file
        net.save_graph(output_html)

        # Inject CSS to center the graph when used in an iframe
        with open(output_html, 'r') as file:
            html_content = file.read()

        # Center the graph by adding custom styles
        centered_style = """
            <style>
              #mynetwork {
                  width: 100%;
                  background-color: #222222;
                  border: 1px solid lightgray;
                  position: relative;
              }

              body, html {
                  margin: 0;
                  padding: 0;
                  overflow: hidden;  /* Prevent scrollbars in the iframe */
                  width: 100%;
                  height: 100%;
              }

              .vis-network {
                  display: flex;
                  justify-content: center;  /* Center horizontally */
                  align-items: center;      /* Center vertically */
              }
            </style>
        """
        custom_script = """
    <script type="text/javascript">
        window.addEventListener('message', function(event) {
            if (event.data && event.data.agentName) {
                const agentName = event.data.agentName;

                // Directly compare agent name with the node ID
                const matchingNode = nodes.get().find(node => node.id === agentName);

                if (matchingNode) {
                    // Reset the color of all nodes to default
                    nodes.forEach(function(node) {
                        nodes.update({ id: node.id, color: '#97c2fc' });
                    });

                    // Highlight the matched node
                    nodes.update({ id: matchingNode.id, color: '#ff6347' });
                }
            }
        });
    </script>
        """

        # Inject the custom style to center the graph in the iframe
        html_content = html_content.replace("</head>", f"{centered_style}</head>")
        html_content = html_content.replace("</body>", f"{custom_script}</body>")

        # Write the modified content back to the HTML file
        with open(output_html, 'w') as file:
            file.write(html_content)

    def create_agent_diagram_from_hocon(self, hocon_file, output_html):
        # Load the HOCON configuration
        agent_data = ConfigFactory.parse_file(hocon_file)

        # Parse the agents and tools from the HOCON data
        agent_graph = self.parse_agent_definitions(agent_data)

        # Create an interactive agent graph and save it as a web page
        if output_html is None:
            # Get the file name without extension
            file_name = Path(hocon_file).stem
            # Generate the output file name
            output_html = str(PATH_TO_STATIC / Path(f"{file_name}.html"))
            print(f"Output file not specified. Saving to {output_html}")

        # Creating the graph will also create the `lib` directory.
        # Make sure it gets created in the **same** directory as the output_html file (e.g. the `static` directory).
        cwd = os.getcwd()
        try:
            static_dir = os.path.dirname(output_html)
            os.chdir(static_dir)
            self.create_interactive_agent_graph(agent_graph, output_html)
        finally:
            os.chdir(cwd)

    def parse_args(self):
        """
        Parse command line arguments into member variables
        """
        arg_parser = argparse.ArgumentParser(
            description="Builds a .html file containing an interactive agent network diagram "
                        "from a .hocon file containing agent definitions and connections."
        )
        arg_parser.add_argument("-i", "--input_file", type=str, default=None, required=True,
                                help="Path to the input .hocon file containing the agent network definition")
        arg_parser.add_argument("-o", "--output_file", type=str, default=None, required=False,
                                help="Path to the file to generate: an .html file that will contain the interactive"
                                     " agent network diagram. If not specified, the output file will be saved"
                                     "in the 'static' folder with the same name as the input file and .html extension.")
        self.args = arg_parser.parse_args()

    def main(self):
        self.parse_args()
        self.create_agent_diagram_from_hocon(hocon_file=self.args.input_file,
                                             output_html=self.args.output_file)


if __name__ == '__main__':
    DiagramBuilder().main()
